@@ -1,5 +1,7 @@
 # -*- coding: utf-8 -*-
 
+from __future__ import absolute_import
+
 import threading
 import operator
 import unittest
@@ -31,31 +33,48 @@ class SetupLuaRuntimeMixin(object):
 
 
 class TestLuaRuntimeRefcounting(unittest.TestCase):
-    def test_runtime_cleanup(self):
-        lua = lupa.LuaRuntime()
-        lua_table = lua.eval('{1,2,3,4}')
-        del lua
-        self.assertEqual(1, lua_table[1])
-        del lua_table
-
-    def test_refcycles(self):
-        def make_refcycle(runtime):
-            def use_runtime():
-                return runtime.eval('1+1')
-            runtime.globals()['use_runtime'] = use_runtime
-            runtime.globals()['some_leaky_data'] = list(range(1000))
-
+    def _run_gc_test(self, run_test):
         gc.collect()
         old_count = len(gc.get_objects())
-        for i in range(1000):
-            lua = lupa.LuaRuntime()
-            make_refcycle(lua)
-            self.assertEqual(2, lua.eval('use_runtime()'))
-            del lua, i
+        i = None
+        for i in range(100):
+            run_test()
+        del i
         gc.collect()
         new_count = len(gc.get_objects())
         self.assertEqual(old_count, new_count)
 
+    def test_runtime_cleanup(self):
+        def run_test():
+            lua = lupa.LuaRuntime()
+            lua_table = lua.eval('{1,2,3,4}')
+            del lua
+            self.assertEqual(1, lua_table[1])
+
+        self._run_gc_test(run_test)
+
+    def test_pyfunc_refcycle(self):
+        def make_refcycle():
+            def use_runtime():
+                return lua.eval('1+1')
+
+            lua = lupa.LuaRuntime()
+            lua.globals()['use_runtime'] = use_runtime
+            self.assertEqual(2, lua.eval('use_runtime()'))
+
+        self._run_gc_test(make_refcycle)
+
+    def test_attrgetter_refcycle(self):
+        def make_refcycle():
+            def get_attr(obj, name):
+                lua.eval('1+1')  # create ref-cycle with runtime
+                return 23
+
+            lua = lupa.LuaRuntime(attribute_handlers=(get_attr, None))
+            assert lua.eval('python.eval.huhu') == 23
+
+        self._run_gc_test(make_refcycle)
+
 
 class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_eval(self):
@@ -418,6 +437,128 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
 
         self.assertEqual(6, len(table))
 
+    def test_table_from_dict(self):
+        table = self.lua.table_from({"foo": 1, "bar": 20, "baz": "spam"})
+        self.assertEqual(     1, table['foo'])
+        self.assertEqual(    20, table['bar'])
+        self.assertEqual("spam", table['baz'])
+
+        self.assertEqual(0, len(table))
+
+    def test_table_from_int_keys(self):
+        table = self.lua.table_from({1: 5, 2: 10, "foo": "bar"})
+        self.assertEqual(5, table[1])
+        self.assertEqual(10, table[2])
+        self.assertEqual("bar", table["foo"])
+
+        self.assertEqual(2, len(table))
+
+    # def test_table_from_obj_keys(self):
+    #     key = object()
+    #     table = self.lua.table_from({key: "foo"})
+    #     self.assertEqual("foo", table[key])
+    #
+    #     self.assertEqual(0, len(table))
+
+    def test_table_from_list(self):
+        table = self.lua.table_from([1,2,5,6])
+        self.assertEqual(1, table[1])
+        self.assertEqual(2, table[2])
+        self.assertEqual(5, table[3])
+        self.assertEqual(6, table[4])
+
+        self.assertEqual(4, len(table))
+
+    def test_table_from_iterable(self):
+        it = (x for x in range(3))
+        table = self.lua.table_from(it)
+        self.assertEqual(0, table[1])
+        self.assertEqual(1, table[2])
+        self.assertEqual(2, table[3])
+
+        self.assertEqual(3, len(table))
+
+    def test_table_from_multiple_dicts(self):
+        table = self.lua.table_from({"a": 1, "b": 2}, {"c": 3, "b": 4})
+        self.assertEqual(1, table["a"])
+        self.assertEqual(4, table["b"])
+        self.assertEqual(3, table["c"])
+
+        self.assertEqual(0, len(table))
+
+    def test_table_from_dicts_and_lists(self):
+        dct, lst = {"a": 1, "b": 2}, ["foo", "bar"]
+        for args in [[dct, lst], [lst, dct]]:
+            table = self.lua.table_from(*args)
+            self.assertEqual(1, table["a"])
+            self.assertEqual(2, table["b"])
+            self.assertEqual("foo", table[1])
+            self.assertEqual("bar", table[2])
+
+            self.assertEqual(2, len(table))
+
+    def test_table_from_multiple_lists(self):
+        table = self.lua.table_from(["foo", "bar"], ["egg", "spam"])
+        self.assertEqual("foo", table[1])
+        self.assertEqual("bar", table[2])
+        self.assertEqual("egg", table[3])
+        self.assertEqual("spam", table[4])
+
+        self.assertEqual(4, len(table))
+
+    def test_table_from_bad(self):
+        self.assertRaises(TypeError, self.lua.table_from, 5)
+        self.assertRaises(TypeError, self.lua.table_from, None)
+        self.assertRaises(TypeError, self.lua.table_from, {"a": 5}, 123)
+
+    # def test_table_from_nested(self):
+    #     table = self.lua.table_from({"obj": {"foo": "bar"}})
+    #     lua_type = self.lua.eval("type")
+    #     self.assertEqual(lua_type(table["obj"]), "table")
+
+    def test_table_from_table(self):
+        table1 = self.lua.eval("{3, 4, foo='bar'}")
+        table2 = self.lua.table_from(table1)
+
+        self.assertEqual(3, table2[1])
+        self.assertEqual(4, table2[2])
+        self.assertEqual("bar", table2["foo"])
+
+        # data should be copied
+        table2["foo"] = "spam"
+        self.assertEqual("spam", table2["foo"])
+        self.assertEqual("bar", table1["foo"])
+
+    def test_table_from_table_iter(self):
+        table1 = self.lua.eval("{3, 4, foo='bar'}")
+        table2 = self.lua.table_from(table1.keys())
+
+        self.assertEqual(len(table2), 3)
+        self.assertEqual(list(table2.keys()), [1, 2, 3])
+        self.assertEqual(set(table2.values()), set([1, 2, "foo"]))
+
+    def test_table_from_table_iter_indirect(self):
+        table1 = self.lua.eval("{3, 4, foo='bar'}")
+        table2 = self.lua.table_from(k for k in table1.keys())
+
+        self.assertEqual(len(table2), 3)
+        self.assertEqual(list(table2.keys()), [1, 2, 3])
+        self.assertEqual(set(table2.values()), set([1, 2, "foo"]))
+
+    # FIXME: it segfaults
+    # def test_table_from_generator_calling_lua_functions(self):
+    #     func = self.lua.eval("function (obj) return obj end")
+    #     table = self.lua.table_from(func(obj) for obj in ["foo", "bar"])
+    #
+    #     self.assertEqual(len(table), 2)
+    #     self.assertEqual(set(table.values()), set(["foo", "bar"]))
+
+    def test_table_contains(self):
+        table = self.lua.eval("{foo=5}")
+        self.assertTrue("foo" in table)
+        self.assertFalse("bar" in table)
+        self.assertFalse(5 in table)
+
     def test_getattr(self):
         stringlib = self.lua.eval('string')
         self.assertEqual('abc', stringlib.lower('ABC'))
@@ -633,6 +774,26 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         function = lua.eval('function(obj) return obj.a end')
         self.assertEqual(function(x), 1)
 
+    def test_lua_type(self):
+        x = self.lua.eval('{}')
+        self.assertEqual('table', lupa.lua_type(x))
+
+        x = self.lua.eval('function() return 1 end')
+        self.assertEqual('function', lupa.lua_type(x))
+
+        x = self.lua.eval('function() coroutine.yield(1) end')
+        self.assertEqual('function', lupa.lua_type(x))
+        self.assertEqual('thread', lupa.lua_type(x.coroutine()))
+
+        self.assertEqual(None, lupa.lua_type(1))
+        self.assertEqual(None, lupa.lua_type(1.1))
+        self.assertEqual(None, lupa.lua_type('abc'))
+        self.assertEqual(None, lupa.lua_type({}))
+        self.assertEqual(None, lupa.lua_type([]))
+        self.assertEqual(None, lupa.lua_type(lupa))
+        self.assertEqual(None, lupa.lua_type(lupa.lua_type))
+
+
 class TestAttributeHandlers(unittest.TestCase):
     def setUp(self):
         self.lua = lupa.LuaRuntime()
@@ -808,7 +969,12 @@ class TestAttributeHandlers(unittest.TestCase):
 
 class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_explicit_python_function(self):
-        lua_func = self.lua.eval('function(func) t = {1, 5, 2, 4, 3}; table.sort(t, python.as_function(func)); return t end')
+        lua_func = self.lua.eval(
+            'function(func)'
+            ' t = {1, 5, 2, 4, 3};'
+            ' table.sort(t, python.as_function(func));'
+            ' return t end')
+
         def compare(a, b):
             return a < b
         self.assertEqual([1, 2, 3, 4, 5], list(lua_func(compare)))
@@ -821,6 +987,7 @@ class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_pyobject_wrapping_callable(self):
         lua_type = self.lua.eval('type')
         lua_get_call = self.lua.eval('function(obj) return getmetatable(obj).__call end')
+
         class Callable(object):
             def __call__(self): pass
             def __getitem__(self, item): pass
@@ -831,6 +998,7 @@ class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_pyobject_wrapping_getitem(self):
         lua_type = self.lua.eval('type')
         lua_get_index = self.lua.eval('function(obj) return getmetatable(obj).__index end')
+
         class GetItem(object):
             def __getitem__(self, item): pass
 
@@ -840,6 +1008,7 @@ class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_pyobject_wrapping_getattr(self):
         lua_type = self.lua.eval('type')
         lua_get_index = self.lua.eval('function(obj) return getmetatable(obj).__index end')
+
         class GetAttr(object):
             pass
 
@@ -1764,6 +1933,170 @@ class TestMethodCall(unittest.TestCase):
         self.assertRaises(TypeError, self.lua.eval, "bound1(x)")
 
 
+################################################################################
+# tests for the lupa.unpacks_lua_table and lupa.unpacks_lua_table_method
+# decorators
+
+@lupa.unpacks_lua_table
+def func_1(x):
+    return ("x=%s" % (x, ))
+
+
+@lupa.unpacks_lua_table
+def func_2(x, y):
+    return ("x=%s, y=%s" % (x, y))
+
+
+@lupa.unpacks_lua_table
+def func_3(x, y, z='default'):
+    return ("x=%s, y=%s, z=%s" % (x, y, z))
+
+
+class MyCls_1(object):
+    @lupa.unpacks_lua_table_method
+    def meth(self, x):
+        return ("x=%s" % (x,))
+
+
+class MyCls_2(object):
+    @lupa.unpacks_lua_table_method
+    def meth(self, x, y):
+        return ("x=%s, y=%s" % (x, y))
+
+
+class MyCls_3(object):
+    @lupa.unpacks_lua_table_method
+    def meth(self, x, y, z='default'):
+        return ("x=%s, y=%s, z=%s" % (x, y, z))
+
+
+class KwargsDecoratorTest(SetupLuaRuntimeMixin, unittest.TestCase):
+
+    def __init__(self, *args, **kwargs):
+        super(KwargsDecoratorTest, self).__init__(*args, **kwargs)
+        self.arg1 = func_1
+        self.arg2 = func_2
+        self.arg3 = func_3
+
+    def assertResult(self, f, call_txt, res_txt):
+        lua_func = self.lua.eval("function (f) return f%s end" % call_txt)
+        self.assertEqual(lua_func(f), res_txt)
+
+    def assertIncorrect(self, f, call_txt):
+        lua_func = self.lua.eval("function (f) return f%s end" % call_txt)
+        self.assertRaises(TypeError, lua_func, f)
+
+    def test_many_args(self):
+        self.assertResult(self.arg2, "{x=1, y=2}", "x=1, y=2")
+        self.assertResult(self.arg2, "{x=2, y=1}", "x=2, y=1")
+        self.assertResult(self.arg2, "{y=1, x=2}", "x=2, y=1")
+        self.assertResult(self.arg2, "(1, 2)",     "x=1, y=2")
+
+    def test_single_arg(self):
+        self.assertResult(self.arg1, "{x=1}", "x=1")
+        self.assertResult(self.arg1, "(1)", "x=1")
+        self.assertResult(self.arg1, "(nil)", "x=None")
+
+    def test_defaults(self):
+        self.assertResult(self.arg3, "{x=1, y=2}", "x=1, y=2, z=default")
+        self.assertResult(self.arg3, "{x=1, y=2, z=3}", "x=1, y=2, z=3")
+
+    def test_defaults_incorrect(self):
+        self.assertIncorrect(self.arg3, "{x=1, z=3}")
+
+    def test_kwargs_unknown(self):
+        self.assertIncorrect(self.arg2, "{x=1, y=2, z=3}")
+        self.assertIncorrect(self.arg2, "{y=2, z=3}")
+        self.assertIncorrect(self.arg1, "{x=1, y=2}")
+
+    def test_posargs_bad(self):
+        self.assertIncorrect(self.arg1, "(1,2)")
+        self.assertIncorrect(self.arg1, "()")
+
+    def test_posargs_kwargs(self):
+        self.assertResult(self.arg2, "{5, y=6}", "x=5, y=6")
+        self.assertResult(self.arg2, "{y=6, 5}", "x=5, y=6")
+        self.assertResult(self.arg2, "{5, [2]=6}", "x=5, y=6")
+        self.assertResult(self.arg2, "{[1]=5, [2]=6}", "x=5, y=6")
+        self.assertResult(self.arg2, "{[1]=5, y=6}", "x=5, y=6")
+
+        self.assertResult(self.arg3, "{x=5, y=6, z=8}", "x=5, y=6, z=8")
+        self.assertResult(self.arg3, "{5, y=6, z=8}", "x=5, y=6, z=8")
+        self.assertResult(self.arg3, "{5, y=6}", "x=5, y=6, z=default")
+        self.assertResult(self.arg3, "{5, 6}", "x=5, y=6, z=default")
+        self.assertResult(self.arg3, "{5, 6, 7}", "x=5, y=6, z=7")
+        self.assertResult(self.arg3, "{z=7, 5, 6}", "x=5, y=6, z=7")
+
+    def test_posargs_kwargs_bad(self):
+        self.assertIncorrect(self.arg2, "{5, y=6, z=7}")
+        self.assertIncorrect(self.arg2, "{5, [3]=6}")
+        self.assertIncorrect(self.arg2, "{x=5, [2]=6}")  # I guess it's ok to reject this
+
+        self.assertIncorrect(self.arg3, "{5, z=7}")
+        self.assertIncorrect(self.arg3, "{5}")
+
+    def test_posargs_nil(self):
+        self.assertResult(self.arg3, "(5, nil, 6)", "x=5, y=None, z=6")
+
+    def test_posargs_nil_first(self):
+        self.assertResult(self.arg3, "(nil, nil, 6)", "x=None, y=None, z=6")
+
+    def test_posargs_nil_last(self):
+        self.assertResult(self.arg3, "(5, nil, nil)", "x=5, y=None, z=None")
+
+    def test_posargs_kwargs_python_none_last(self):
+        self.assertResult(self.arg3, "{5, python.none, python.none}", "x=5, y=None, z=None")
+
+    def test_posargs_python_none_last(self):
+        self.assertResult(self.arg3, "(5, python.none, python.none)", "x=5, y=None, z=None")
+
+    def test_posargs_kwargs_python_none_some(self):
+        self.assertResult(self.arg3, "{python.none, y=python.none, z=6}", "x=None, y=None, z=6")
+
+    def test_posargs_kwargs_python_none_all(self):
+        self.assertResult(self.arg3, "{x=python.none, y=python.none}", "x=None, y=None, z=default")
+
+    # -------------------------------------------------------------------------
+    # The following examples don't work as a Python programmer would expect
+    # them to:
+
+    # def test_posargs_kwargs_nil_last(self):
+    #     self.assertResult(self.arg3, "{5, nil, nil}", "x=5, y=None, z=None")
+    #
+    # def test_posargs_kwargs_nil_some(self):
+    #     self.assertResult(self.arg3, "{nil, y=nil, z=6}", "x=None, y=None, z=6")
+    #
+    # def test_posargs_kwargs_nil_all(self):
+    #     self.assertResult(self.arg3, "{x=nil, y=nil}", "x=None, y=None, z=default")
+
+    # -------------------------------------------------------------------------
+    # These tests pass in Lua 5.2 but fail in LuaJIT:
+
+    # def test_posargs_kwargs_nil(self):
+    #     self.assertResult(self.arg3, "{5, nil, 6}", "x=5, y=None, z=6")
+    #
+    # def test_posargs_kwargs_nil_first(self):
+    #     self.assertResult(self.arg3, "{nil, nil, 6}", "x=None, y=None, z=6")
+
+
+
+class MethodKwargsDecoratorTest(KwargsDecoratorTest):
+
+    def __init__(self, *args, **kwargs):
+        super(MethodKwargsDecoratorTest, self).__init__(*args, **kwargs)
+        self.arg1 = MyCls_1()
+        self.arg2 = MyCls_2()
+        self.arg3 = MyCls_3()
+
+    def assertResult(self, f, call_txt, res_txt):
+        lua_func = self.lua.eval("function (obj) return obj:meth%s end" % call_txt)
+        self.assertEqual(lua_func(f), res_txt)
+
+    def assertIncorrect(self, f, call_txt):
+        lua_func = self.lua.eval("function (obj) return obj:meth%s end" % call_txt)
+        self.assertRaises(TypeError, lua_func, f)
+
+
 ################################################################################
 # tests for the FastRLock implementation
 
